/* This file has been automatically generated.  Do not edit.  */
#ifndef _CADO_PATHS_H
#define _CADO_PATHS_H

#include <config.h>

/* Spool directory path */
#define SPOOL_DIR CADO_SPOOL_DIR

/* Cado temporary exe directory path */
#define CADO_EXE_DIR  "/tmp"

#endif /* _SCADO_PATHS_H */
